import { AliasMapFactory } from "./alias-map.factory";
var Transformer = (function () {
    /**
     *
     * @param transformersMap
     * @param withRelations
     */
    function Transformer(transformersMap, withRelations) {
        if (withRelations === void 0) { withRelations = true; }
        this.transformersMap = transformersMap;
        this.withRelations = withRelations;
        // If aliases present then convert them to aliases map.
        var aliases = this.aliases();
        if (aliases)
            this.aliasesMap = Transformer.createAliasesMap(aliases);
    }
    /**
     * Create aliases map.
     *
     * @param aliases
     * @returns {AliasMap}
     */
    Transformer.createAliasesMap = function (aliases) {
        return new AliasMapFactory(aliases).create();
    };
    /**
     * Transform response data to model entity.
     *
     * @param response
     * @returns {Model}
     */
    Transformer.prototype.call = function (response) {
        this.response = response;
        // List
        if (response.data && response.data instanceof Array)
            return this.response.data.map(this.parse, this);
        else if (response.data)
            return this.parse(response.data);
        else if (response.attributes)
            return this.parse(response);
    };
    /**
     * Parse main model data.
     *
     * @param data
     * @returns {Model}
     */
    Transformer.prototype.parse = function (data) {
        var model = this.transform(data);
        if (this.withRelations && 'relationships' in data && this.transformersMap)
            this.parseRelations(data.relationships, model);
        return model;
    };
    /**
     * Parse relations list.
     *
     * @param relations
     * @param model
     */
    Transformer.prototype.parseRelations = function (relations, model) {
        var keys = Object.getOwnPropertyNames(relations);
        for (var _i = 0, keys_1 = keys; _i < keys_1.length; _i++) {
            var key = keys_1[_i];
            var relation = relations[key].data;
            if (relation instanceof Array) {
                for (var _a = 0, relation_1 = relation; _a < relation_1.length; _a++) {
                    var _relation = relation_1[_a];
                    this.parseRelation(_relation, model, true);
                }
            }
            else
                this.parseRelation(relation, model);
        }
    };
    /**
     * Parse the relation.
     *
     * @param data
     * @param model
     * @param isArray
     */
    Transformer.prototype.parseRelation = function (data, model, isArray) {
        if (isArray === void 0) { isArray = false; }
        var alias = this.findAlias(data.type);
        var type = data.type;
        var include = this.findInclude(type, data.id);
        this.parseInclude(include, model, alias, isArray);
    };
    /**
     * Find an alias for model type.
     *
     * @param type
     * @returns {undefined|string}
     */
    Transformer.prototype.findAlias = function (type) {
        if (this.aliasesMap)
            return this.aliasesMap.get(type);
    };
    /**
     * Find model include data.
     *
     * @param type
     * @param id
     * @returns {any}
     */
    Transformer.prototype.findInclude = function (type, id) {
        for (var _i = 0, _a = this.response.included; _i < _a.length; _i++) {
            var include = _a[_i];
            if (include.type == type && include.id == id)
                return include;
        }
    };
    /**
     * Parse model include data.
     *
     * @param include
     * @param model
     * @param alias
     * @param isArray
     */
    Transformer.prototype.parseInclude = function (include, model, alias, isArray) {
        if (isArray === void 0) { isArray = false; }
        if (!include.type && !alias)
            throw new Error('Include item type or alias must be set. Please check your backend API response.');
        var type = include.type;
        var transformer = this.transformersMap.get(alias || type);
        var includeModel = transformer ? transformer.transform(include) : include;
        // Parse sub include relations if exists
        if (include.relationships)
            this.parseRelations(include.relationships, includeModel);
        if (isArray) {
            if (!model[type])
                model[type] = [];
            model[type].push(includeModel);
        }
        else
            model[type] = includeModel;
    };
    return Transformer;
}());
export { Transformer };
//# sourceMappingURL=transformer.js.map