import {IApiService} from "./iapi-service";
import {Http, RequestOptions, RequestOptionsArgs, Response} from "@angular/http";
import {ApiResponse} from "./api-response";
import {Observable} from 'rxjs/Observable';
import 'rxjs/add/operator/catch';
import 'rxjs/add/operator/map';
import {ITransformer} from "./itransformer";
import {Search} from "./search";
import {Order} from "./order";
import {Filter} from "./filter";
import {Factory} from "./factory";
import {ApiCollection} from "./api.collection";
import {SearchFactory} from "./search.factory";
import {OrderFactory} from "./order.factory";
import {FilterFactory} from "./filter.factory";
import {Relation} from "./relation";
import {RelationCollection} from "./relation.collection";
import {ErrorObservable} from "rxjs/observable/ErrorObservable";
import {IEntity} from "./ientity";


export abstract class ApiService<Model extends IEntity> implements IApiService<Model> {

    public static readonly urlSearchQueryParam = 'search';
    public static readonly urlSearchFieldsParam = 'searchFields';
    public static readonly urlOrderParam = 'orderBy';
    public static readonly urlSortParam = 'sortedBy';
    public static readonly urlFilterParam = 'filter';
    public static readonly urlIncludeParam = 'include';

    constructor(private http: Http) {}

    protected extractData(res: Response): any {
        return new ApiResponse(this, res.json() || {});
    }

    protected handleError(error: Response | any): ErrorObservable {
        let errMsg: string;
        if (error instanceof Response) {
            const body = error.json() || '';
            const err = body.error || JSON.stringify(body);
            errMsg = `${error.status} - ${error.statusText || ''} ${err}`;
        } else {
            errMsg = error.message ? error.message : error.toString();
        }
        console.error(errMsg);
        return Observable.throw(errMsg);
        // return 'custom error';
    }

    public abstract transformer(): ITransformer<Model>;

    /*****************************************************************************************************************
     *      Http client methods aliases.
     */

    public get(url: string,
               search?: Search[], orders?: Order[], filters?: Filter[], relations?: Relation[],
               options?: RequestOptionsArgs): Observable<ApiResponse<Model>> {

        if (!options)
            options = new RequestOptions();

        const params: any = {};

        if (search) {
            const factory: Factory<ApiCollection<Search>> = new SearchFactory(search);

            params[ApiService.urlSearchQueryParam] = factory.create(SearchFactory.TYPE_QUERY).toUrl();
            params[ApiService.urlSearchFieldsParam] = factory.create(SearchFactory.TYPE_FIELDS).toUrl();
        }

        if (orders) {
            const factory: Factory<ApiCollection<Order>> = new OrderFactory(orders);

            params[ApiService.urlOrderParam] = factory.create(OrderFactory.TYPE_QUERY).toUrl();
            params[ApiService.urlSortParam] = factory.create(OrderFactory.TYPE_FIELDS).toUrl();
        }

        if (filters)
            params[ApiService.urlFilterParam] = new FilterFactory(filters).create(FilterFactory.TYPE_FIELDS).toUrl();

        if (relations)
            params[ApiService.urlIncludeParam] = new RelationCollection(relations).toUrl();

        options.params = Object.assign({}, params, options.params);

        // hack for using service context instance
        const self = this;
        return this.http.get(url, options)
            .map((res: Response) => { return self.extractData(res) })
            .catch((error: Response | any) => { return self.handleError(error) });
    }

    public post(url: string, body: Model | any, options?: RequestOptionsArgs): Observable<ApiResponse<Model>> {
        // hack for using service context instance
        const self = this;
        return this.http.post(url, body, options)
            .map((res: Response) => { return self.extractData(res) })
            .catch((error: Response | any) => { return self.handleError(error) });
    }

    public put(url: string, body: Model | any, options?: RequestOptionsArgs): Observable<ApiResponse<Model>> {
        const self = this;
        return this.http.put(url, body, options)
            .map(this.extractData)
            .catch((error: Response | any) => { return self.handleError(error) });
    }

    public patch(url: string, body: Model | any, options?: RequestOptionsArgs): Observable<ApiResponse<Model>> {
        const self = this;
        return this.http.patch(url, body, options)
            .map(this.extractData)
            .catch((error: Response | any) => { return self.handleError(error) });
    }

    public delete(url: string, options?: RequestOptionsArgs): Observable<ApiResponse<Model>> {
        const self = this;
        return this.http.delete(url, options)
            .map(this.extractData)
            .catch((error: Response | any) => { return self.handleError(error) });
    }

    public head(url: string, options?: RequestOptionsArgs): Observable<Response> {
        return this.http.head(url, options);
    }

    public options(url: string, options?: RequestOptionsArgs): Observable<Response> {
        return this.http.options(url, options);
    }

}