import {ITransformerMap} from "./itransformer.map";

import {Aliases} from "./aliases";
import {AliasMap} from "./alias.map";

import {AliasMapFactory} from "./alias-map.factory";
import {ITransformer} from "./itransformer";

import {IEntity} from "./ientity";

export abstract class Transformer<Model> implements ITransformer<Model> {
    private response: any;
    private aliasesMap: AliasMap | undefined;

    /**
     *
     * @param transformersMap
     * @param withRelations
     */
    constructor(private transformersMap: ITransformerMap,
                private withRelations: boolean = true)
    {
        // If aliases present then convert them to aliases map.
        const aliases: Aliases | undefined = this.aliases();
        if (aliases)
            this.aliasesMap = Transformer.createAliasesMap(aliases);
    }

    /**
     * Create aliases map.
     *
     * @param aliases
     * @returns {AliasMap}
     */
    private static createAliasesMap(aliases: Aliases): AliasMap {
        return new AliasMapFactory(aliases).create();
    }

    /**
     * Aliases map.
     * Using for setup transformer aliases for different entity types.
     */
    public abstract aliases(): Aliases | undefined;

    /**
     * Transform method.
     *
     * @param data
     */
    public abstract transform(data: any): Model;

    /**
     * Transform response data to model entity.
     *
     * @param response
     * @returns {Model}
     */
    public call(response: any): Model | Model[] | undefined {
        this.response = response;

        // List
        if (response.data && response.data instanceof Array)
            return this.response.data.map(this.parse, this);
        // Item
        else if (response.data)
            return this.parse(response.data);
        // Included item
        else if (response.attributes)
            return this.parse(response);
    }

    /**
     * Parse main model data.
     *
     * @param data
     * @returns {Model}
     */
    private parse(data: any): Model {
        const model: Model = this.transform(data);

        if (this.withRelations && 'relationships' in data && this.transformersMap)
            this.parseRelations(data.relationships, model);

        return model;
    }

    /**
     * Parse relations list.
     *
     * @param relations
     * @param model
     */
    private parseRelations(relations: any, model: Model): void {
        const keys: string[] = Object.getOwnPropertyNames(relations);

        for (let key of keys) {
            const relation: any = relations[key].data;

            if (relation instanceof Array) {
                for (let _relation of relation)
                    this.parseRelation(_relation, model, true);
            } else
                this.parseRelation(relation, model);
        }
    }

    /**
     * Parse the relation.
     *
     * @param data
     * @param model
     * @param isArray
     */
    private parseRelation(data: any, model: Model, isArray: boolean = false): void {
        const alias: string | undefined = this.findAlias(data.type);
        const type: string = data.type;

        const include: any = this.findInclude(type, data.id);

        this.parseInclude(include, model, alias, isArray);
    }

    /**
     * Find an alias for model type.
     *
     * @param type
     * @returns {undefined|string}
     */
    private findAlias(type: string): string | undefined {
        if (this.aliasesMap)
            return this.aliasesMap.get(type);
    }

    /**
     * Find model include data.
     *
     * @param type
     * @param id
     * @returns {any}
     */
    private findInclude(type: string, id: number | string): any | undefined {
        for (let include of this.response.included) {
            if (include.type == type && include.id == id)
                return include;
        }
    }

    /**
     * Parse model include data.
     *
     * @param include
     * @param model
     * @param alias
     * @param isArray
     */
    private parseInclude(include: any, model: Model | any, alias?: string, isArray: boolean = false): void {
        if (!include.type && !alias)
            throw new Error('Include item type or alias must be set. Please check your backend API response.');

        const type = include.type;
        const transformer: ITransformer<IEntity> = this.transformersMap.get(alias || type);

        const includeModel = transformer ? transformer.transform(include) : include;

        // Parse sub include relations if exists
        if (include.relationships)
            this.parseRelations(include.relationships, includeModel);

        if (isArray) {
            if (!model[type])
                model[type] = [];

            model[type].push(includeModel);
        } else
            model[type] = includeModel;
    }
}